from browser import (
    bind, document, window,
    # So we can log like in js by doing `browser_console.log(...)`
    console as browser_console,
)
import sys
import tb as traceback
import javascript
from re import compile, finditer
from browser.session_storage import storage

#############
# CONSTANTS #
#############
INIT_COMMENTS = """# Les zones grisées ne doivent pas être modifiées !\n"""
INIT_CONSOLE_MSG = """Appuyer sur le bouton "LANCER" ci-dessus pour exécuter \
le code présent dans l'éditeur. Puis taper une commande ci-dessous et appuyer \
sur la touche Entrée pour l'exécuter."""

#########
# UTILS #
#########
def set_data(question_number):
    """returns if editor should be on "edit mode" or locked.
    not very intuitive but definitely the most efficient."""
    response_body = javascript.this().response_body[question_number]
    if isinstance(response_body, str):
        data_temp = javascript.JSON.parse(response_body)
    else:
        data_temp = response_body.to_dict()
    if is_exercise_object(response_body):
        data_temp['initCode'] = INIT_COMMENTS + multi_line_disable(tabs_to_space(data_temp['initCode']))
        display_only = False
    else:
        data_temp['initCode'] = multi_line_disable(tabs_to_space(data_temp['student_code']))
        del data_temp['student_code']
        display_only = True
    if "data" in globals():
        globals()["data"].append(data_temp)
    else:
        globals()["data"] = [data_temp]

    return display_only


def trigger_disable(editor):
    javascript.this().disable_ranges(editor)


def trigger_enable(editor):
    return javascript.this().enable_ranges(editor)


def toggle_lock(editor_container, toggler):
    editor = get_editor(editor_container)
    editor.setOptions({
        "dragEnabled": not toggler,
        "readOnly": toggler
    })


def get_question_number(ele_start):
    current = ele_start
    while current.parent and "data-editornbr" not in current.attrs:
        current = current.parent
    return int(current.attrs["data-editornbr"])


def get_question_elements(ele_start):
    if isinstance(ele_start, int):
        return editors[ele_start], consoles[ele_start]
    else:
        question_number = get_question_number(ele_start)
        return editors[question_number], consoles[question_number]


def get_console_output(ele_start):
    if isinstance(ele_start, int):
        return consoles[ele_start]
    else:
        question_number = get_question_number(ele_start)
        return consoles[question_number]


def get_editor(ele_start):
    if isinstance(ele_start, int):
        return editors[ele_start]
    else:
        question_number = get_question_number(ele_start)
        return editors[question_number]


def tabs_to_space(code):
    return code.replace("\t", "    ")


def is_exercise_object(data):
    return "initCode" in data


# Transform traceback as understandable traceback for students
def transform_error(error_type, local_traceback, lines_added=[], msg=None):
    if error_type in ["NameError", "UnboundLocalError"]:
        output_msg = ""
    if error_type == "SyntaxError":
        output_msg = "Erreur de syntaxe détectée"
    elif error_type == "TypeError":
        output_msg = "Erreur de type détectée"
    elif error_type == "ValueError":
        output_msg = "Erreur de valeur détectée"
        if msg and msg[0] == "BOUCLE INFINIE":
            return "Une boucle infinie a été détectée."
    elif error_type == "ArithmeticError":
        output_msg = "Erreur de calcul détectée"
    else:
        output_msg = error_type + " détectée"

    regex = compile(r"(line \d+)")
    lineErr = regex.findall(local_traceback.format_exc())
    if len(lineErr):
        line_nbr_err = int(lineErr[len(lineErr) - 1].split(" ")[1])
        # Ignore lines added to prevent infinite loops
        line_nbr = line_nbr_err - len([ln for ln in lines_added if ln <= line_nbr_err])
        output_msg += " à la ligne " + str(line_nbr) + "\n"

    if error_type in ["NameError", "UnboundLocalError"] and msg is not None:
        regex = compile("\'.*\'")
        nameErr = regex.search(msg).group(0)
        output_msg += nameErr + " n'est pas défini"

    return output_msg


def hide_all_alerts(question_nbr):
    alert_name = "load-alert-fail-" + question_nbr
    document[alert_name].style.display = "none"

    alert_name = "load-alert-success-" + question_nbr
    document[alert_name].style.display = "none"

    alert_name = "save-alert-" + question_nbr
    document[alert_name].style.display = "none"


###########
# CLASSES #
###########
class Trace:

    def __init__(self):
        self.buf = ""

    def write(self, data):
        self.buf += str(data)

    def format(self):
        """Remove calls to function in this script from the traceback."""
        lines = self.buf.split("\n")
        stripped = [lines[0]]
        for i in range(1, len(lines), 2):
            if __file__ in lines[i]:
                continue
            stripped += lines[i: i + 2]
        return "\n".join(stripped)


class Console:

    def __init__(self, element, namespace={}):
        self.element = element
        element.bind('keypress', self.keypress)
        element.bind('keydown', self.keydown)
        element.bind('click', self.cursorToEnd)
        self.cursorToEnd()
        self.status = "main"
        self.current = 0
        self.history = []
        self.buffer = ""
        sys.stdout = self
        sys.stderr = self
        self.namespace = namespace
        self.write(INIT_CONSOLE_MSG)

    def clear(self):
        self.buffer = ''
        self.element.value = ''

    def cursorToEnd(self, *args):
        pos = len(self.element.value)
        self.element.setSelectionRange(pos, pos)
        self.element.scrollTop = self.element.scrollHeight

    def keypress(self, event):
        if event.keyCode == 9:  # tab key
            event.preventDefault()
            self.element.value += "    "
        elif event.keyCode == 13:  # return
            src = self.element.value
            if self.status == "main":
                currentLine = src[src.rfind('>>>') + 4:]
            elif self.status == "3string":
                currentLine = src[src.rfind('>>>') + 4:]
                currentLine = currentLine.replace('\n... ', '\n')
            else:
                currentLine = src[src.rfind('...') + 4:]
            if self.status == 'main' and not currentLine.strip():
                self.element.value += '\n>>> '
                event.preventDefault()
                return
            self.element.value += '\n'
            self.history.append(currentLine)
            self.current = len(self.history)
            if self.status == "main" or self.status == "3string":
                try:
                    _ = self.namespace['_'] = eval(currentLine, self.namespace)
                    if _ is not None:
                        self.write(repr(_) + '\n')
                    self.write('>>> ')
                    self.status = "main"
                except IndentationError:
                    self.element.value += '... '
                    self.status = "block"
                except SyntaxError as msg:
                    if str(msg) == 'invalid syntax : triple string end not found' or \
                            str(msg).startswith('Unbalanced bracket'):
                        self.element.value += '... '
                        self.status = "3string"
                    elif str(msg) == 'eval() argument must be an expression':
                        try:
                            exec(currentLine, self.namespace)
                        except Exception:
                            self.print_tb()
                        self.write('>>> ')
                        self.status = "main"
                    elif str(msg) == 'decorator expects function':
                        self.element.value += '... '
                        self.status = "block"
                    else:
                        self.print_tb(is_syntax_error=True)
                        self.element.value += '>>> '
                        self.status = "main"
                except Exception:
                    # the full traceback includes the call to eval(); to
                    # remove it, it is stored in a buffer and the 2nd and 3rd
                    # lines are removed
                    self.print_tb()
                    self.element.value += '>>> '
                    self.status = "main"
            elif currentLine == "":  # end of block
                block = src[src.rfind('>>>') + 4:].splitlines()
                block = [block[0]] + [b[4:] for b in block[1:]]
                block_src = '\n'.join(block)
                # status must be set before executing code in globals()
                self.status = "main"
                try:
                    code_python, _ = prevent_infinite_loops(block_src)
                    _ = exec(code_python, self.namespace)
                    if _ is not None:
                        print(repr(_))
                except Exception:
                    self.print_tb()
                self.write('>>> ')
            else:
                self.element.value += '... '

            self.cursorToEnd()
            event.preventDefault()

    def keydown(self, event):
        if event.keyCode == 37:  # left arrow
            sel = self.get_col()
            if sel < 5:
                event.preventDefault()
                event.stopPropagation()
        elif event.keyCode == 36:  # line start
            pos = self.element.selectionStart
            col = self.get_col()
            self.element.setSelectionRange(pos - col + 4, pos - col + 4)
            event.preventDefault()
        elif event.keyCode == 38:  # up
            if self.current > 0:
                pos = self.element.selectionStart
                col = self.get_col()
                # remove current line
                self.element.value = self.element.value[:pos - col + 4]
                self.current -= 1
                self.element.value += self.history[self.current]
            event.preventDefault()
        elif event.keyCode == 40:  # down
            if self.current < len(self.history) - 1:
                pos = self.element.selectionStart
                col = self.get_col()
                # remove current line
                self.element.value = self.element.value[:pos - col + 4]
                self.current += 1
                self.element.value += self.history[self.current]
            event.preventDefault()
        elif event.keyCode == 8:  # backspace
            src = self.element.value
            lstart = src.rfind('\n')
            if (lstart == -1 and len(src) < 5) or (len(src) - lstart < 6):
                event.preventDefault()
                event.stopPropagation()
        elif event.ctrlKey and event.keyCode == 65:  # ctrl+a
            src = self.element.value
            pos = self.element.selectionStart
            col = self.get_col()
            self.element.setSelectionRange(pos - col + 4, len(src))
            event.preventDefault()
        elif event.keyCode in [33, 34]:  # page up, page down
            event.preventDefault()

    def get_col(self):
        # returns the column num of cursor
        sel = self.element.selectionStart
        lines = self.element.value.split('\n')
        for line in lines[:-1]:
            sel -= len(line) + 1
        return sel

    def print_tb(self, is_syntax_error=False):
        trace = Trace()
        traceback.print_exc(file=trace)
        trace_msg = trace.format()
        regex = compile(r"(\w*):")
        group_to_catch = 0
        if not is_syntax_error:
            group_to_catch += 1
        error_type = regex.findall(trace_msg)[group_to_catch]
        msg_content = None
        if error_type in ["NameError", "UnboundLocalError"]:
            msg_content = trace_msg
        elif 'BOUCLE INFINIE' in trace_msg:
            msg_content = ['BOUCLE INFINIE']

        lines_added = []
        if 'lines_added' in self.namespace:
            lines_added = self.namespace['lines_added']
        self.write(transform_error(error_type, traceback, lines_added=lines_added, msg=msg_content) + "\n")

    def prompt(self):
        if self.element.value and not self.element.value.endswith('\n'):
            self.write('\n')
        self.write(">>> ")
        self.element.focus()

    def prompt_no_focus(self):
        if self.element.value and not self.element.value.endswith('\n'):
            self.write('\n')
        self.write(">>> ")

    def print_on_prompt(self, msg):
        if self.element.value and not self.element.value.endswith('\n'):
            self.write('\n')
        self.write(">>> " + msg)
        self.element.focus()

    def syntax_error(self, args):
        info, filename, lineno, offset, line = args
        print(f"  File {filename}, line {lineno}")
        print("    " + line)
        print("    " + offset * " " + "^")
        print("SyntaxError:", info)

    def write(self, data):
        self.element.value += str(data)


editors = []
consoles = []
saves_found = []


##########
# EVENTS #
##########
def create_editor(editor_number, display_only_mode, numbers_editors_has_display_only):
    """Create a Python code editor with Ace"""
    global editors
    global saves_found

    window.ace.config.set('basePath', '/static/exercise/js/ace')
    editor = document.select(".editor")[editor_number]
    editor = window.ace.edit(editor)
    editor.setHighlightActiveLine(False)
    editor.setBehavioursEnabled(False)
    editor.setTheme("ace/theme/chrome")
    editor.session.setMode("ace/mode/python")
    initCode = globals()["data"][editor_number]["initCode"]

    if not display_only_mode:
        editor.on("change", editor_changed)
        if "saves" in storage:
            saves = javascript.JSON.parse(storage["saves"])
            instance_element_name = "id_instance_" + str(editor_number)
            for saved in saves:
                if instance_element_name in saved:
                    instance_id = document[instance_element_name].attrs["value"]
                    if saved[instance_element_name] == instance_id:
                        editor.setValue(multi_line_disable(saved["student_code"]))
                        saves_found.append(True)
                        break
            else:
                saves_found.append(False)
                editor.setValue(initCode)
        else:
            editor.setValue(initCode)
    else:
        editor.setValue(initCode)
        editor.setOptions({
            "dragEnabled": False,
            "readOnly": True
        })
    trigger_disable(editor)
    editors.append(editor)
    if not window.areDisplayOnly[editor_number]:
        events = [run, undo, redo, save, load]
        for event in events:
            document.select("." + event.__name__)[editor_number - numbers_editors_has_display_only].bind("click", event)
        editor.session.getUndoManager().reset()


def multi_line_disable(code):
    """If a marker is on multiple line, we deconstruct it line by line for ace purposes.
    see disable_ranges in python_editor.js for more info."""
    code_by_line = code.split("\n")
    code_disabled_by_line = []
    is_inside_brackets = False
    for code_line in code_by_line:
        current_line = code_line
        if "%{" in code_line and "}s" not in code_line:
            is_inside_brackets = True
            current_line += "}s"
        if is_inside_brackets and "}s" in code_line:
            is_inside_brackets = False
            current_line = "%{" + current_line
        code_disabled_by_line.append(current_line)

    return "\n".join(code_disabled_by_line)


def editor_changed(*args):
    """Called when the editor content changes."""
    current = document.select(".current")
    if current:
        filename = current[0].text.rstrip("*")
        if open_files[filename]["content"] != editor.getValue():
            if not current[0].text.endswith("*"):
                current[0].text += "*"
        elif current[0].text.endswith("*"):
            current[0].text = current[0].text.rstrip("*")


def save(evt):
    """Save the current script in the session storage"""
    question_nbr = get_question_number(evt.target)
    editor = get_editor(question_nbr)
    instance_name = "id_instance_" + str(question_nbr)
    if "saves" in storage:
        saves = javascript.JSON.parse(storage["saves"])
        for idx, saved in enumerate(saves):
            if instance_name in saved:
                saves[idx]["student_code"] = trigger_enable(editor)
                break
        else:
            code_to_save = {"student_code": trigger_enable(editor),
                            instance_name: document[instance_name].attrs["value"]}
            saves.append(code_to_save)
    else:
        code_to_save = {"student_code": trigger_enable(editor),
                        instance_name: document[instance_name].attrs["value"]}
        saves = [code_to_save]
    storage["saves"] = javascript.JSON.stringify(saves)
    question_nbr_str = str(question_nbr)
    hide_all_alerts(question_nbr_str)
    alert_name = "save-alert-" + question_nbr_str
    document[alert_name].style.display = "block"


@bind(".load", "click")
def load(evt):
    """Load the script in the session storage if exists"""
    question_number = get_question_number(evt.target)
    question_number_str = str(question_number)
    editor = get_editor(question_number)
    fail_name = "load-alert-fail-" + question_number_str
    if "saves" in storage:
        saves = javascript.JSON.parse(storage["saves"])
        instance_element_name = "id_instance_" + question_number_str
        for idx, saved in enumerate(saves):
            if instance_element_name in saved:
                if editor.getValue() != saved["student_code"]:
                    editor.setValue(saved["student_code"])
                    trigger_disable(editor)
                    hide_all_alerts(question_number_str)
                    editor.session.getUndoManager().reset()
                break
        else:
            document[fail_name].style.display = "block"
    else:
        document[fail_name].style.display = "block"


@bind("#reset", "click")
def reset(evt):
    """Reset code to its initial form
    i.e: the init code received from request"""
    question_nbr = get_question_number(evt.target)
    editor, console = get_question_elements(question_nbr)
    init_code = globals()["data"][question_nbr]["initCode"]
    editor.setValue(init_code)
    trigger_disable(editor)
    console.clear()
    editor.session.getUndoManager().reset()


def run(evt):
    """Run the script and start the interactive session in the console
    Reformat exceptions for student comprehension."""
    editor, console = get_question_elements(evt.target)
    sys.stderr = console
    sys.stdout = console
    ns = {}
    editor_number = get_question_number(evt.target)
    if "loadedCode" in globals()["data"][editor_number]:
        loadedCode = globals()["data"][editor_number]["loadedCode"]
    else:
        loadedCode = ""
    exec(loadedCode, ns)
    code_python, lines_added = prevent_infinite_loops(editor.getValue())
    ns['lines_added'] = lines_added
    try:
        exec(code_python, ns)
        # Set console namespace
        console.namespace = ns
    except NameError as msg:
        print(transform_error("NameError", traceback, lines_added, msg=msg.args[0]))
    except UnboundLocalError as msg:
        print(transform_error("UnboundLocalError", traceback, lines_added, msg=msg.args[0]))
    except SyntaxError:
        print(transform_error("SyntaxError", traceback, lines_added))
    except TypeError:
        print(transform_error("TypeError", traceback, lines_added))
    except ValueError as msg:
        print(transform_error("ValueError", traceback, lines_added, msg=msg.args))
    except Exception as e:
        print(e)  # Prevent unhandled exception from locking the console

    console.prompt()


@bind(".redo", "click")
def redo(evt):
    editor = get_editor(evt.target)
    manager = editor.session.getUndoManager()
    if manager.hasRedo():
        editor.redo()
        editor.clearSelection()
        editor.focus()


@bind(".undo", "click")
def undo(evt):
    editor = get_editor(evt.target)
    if editor.session.getUndoManager().hasUndo():
        editor.undo()


def keyup_rename(evt, filename):
    """Handle two special keys, Escape and Enter."""
    if evt.keyCode == 27:  # Escape key: cancel
        evt.target.parent.remove()
        update_filebrowser(filename)
        evt.preventDefault()
        evt.stopPropagation()
    elif evt.keyCode == 13:  # Enter key: same as blur
        rename2(evt, filename)
        evt.preventDefault()
        evt.stopPropagation()


def prevent_infinite_loops(code_python):
    """
    Function transforming student code to avoid infinite loops during execution
    as we can't stop them and they make the browser crash.
    We do this by adding a counter to every while loop, when the counter is too
    high, we raise an exception that will stop the execution.
    :param code_python: a string containing student code
    :return: a couple (string, list) where the string is the safe code and the
        list contains the line number of every lines we added to the student
        code to make it safe.
    """
    # Remove trailing whitespaces or we might miss ":\n" further
    safe_python_code = ''.join([line.rstrip() + '\n' for line in code_python.splitlines()])
    lines_added = []

    # Wanted to ignore all while in quotes but could not as the run takes way
    # too much time then.
    for i, match in reversed(list(enumerate(finditer("while", safe_python_code)))):
        counter_name = "_0x3207a6" + str(i)
        index_start_while = match.start()
        index_end_while = safe_python_code.find(':\n', index_start_while) + 1
        if not index_end_while:
            # TODO: we can't prevent infinite loops when there's something
            # after the ":" such as:
            # `while True:  # a comment`
            # `while True: nb += 1`  (which is correct in python)
            # If someone wants to improve this, feel free to do so, but keep
            # in mind that `#` could be in a string and `:` in list access
            # such as `while l[:5]:`.
            # So this might be harder than you think.
            continue
        while_statement = safe_python_code[index_start_while:index_end_while]

        # Get line number, since we parse in reverse order we don't have to
        # bother about already added lines
        ln = safe_python_code[:index_start_while].count('\n') + 1

        # Find indent length where while is defined
        outer_indent = 0
        while safe_python_code[index_start_while - outer_indent - 1] == " ":
            outer_indent += 1
        outer_indent = outer_indent * " "

        # Find indent length inside the while loop
        inner_indent = 0
        while safe_python_code[index_end_while + inner_indent + 1] == " ":
            inner_indent += 1
        inner_indent = inner_indent * " "

        safe_python_code = (
            "{code_before}"
            "{counter_name} = 0\n"
            "{outer_indent}{while_statement}\n"
            "{inner_indent}{counter_name} += 1\n"
            "{inner_indent}if {counter_name} > 1000000:\n"
            "{inner_indent}    raise ValueError('BOUCLE INFINIE')"
            "{code_after}"
            .format(
                code_before=safe_python_code[:index_start_while],
                code_after=safe_python_code[index_end_while:],
                while_statement=while_statement,
                counter_name=counter_name,
                outer_indent=outer_indent,
                inner_indent=inner_indent,
            )
        )

        lines_added = [l_ + 4 for l_ in lines_added]
        lines_added = [ln, ln + 2, ln + 3, ln + 4] + lines_added

    return safe_python_code, lines_added


def create_function_object(code_python, question_data):
    """Used by create_request, cast student code from string to a function object
    Return None values if the action creates an Exception
        meaning the string cannot be turned into a function object"""
    global_values = dict()
    local_values = dict()

    code_python, _ = prevent_infinite_loops(code_python)

    try:
        exec(code_python, global_values, local_values)
    except Exception:
        return lambda *args: 'None'
    if local_values:
        if "funcName" in question_data:
            funcName = question_data["funcName"]
            if funcName in local_values:
                return local_values[funcName]
            else:
                return lambda *args: 'NoFunc'
        else:
            return list(local_values.values())[4]


def create_request(question_data, code):
    """Used by submit code, transform values and code into outputs for back-end."""
    f = create_function_object(code, question_data)
    values = question_data["inputs"]

    def catch_error(function, inputs):
        try:
            output_student = function(*inputs)
            if output_student == "None" or output_student == "NoFunc":
                return output_student
            else:
                # Answer from student (not "created" by code) should always be in an array
                return [output_student]
        except Exception:
            return 'None'

    return {
        "outputs": [catch_error(f, value) for value in values],
        "student_code": code
    }


def submit_code(evt):
    """Triggered before submit exercise"""

    # Prevent print to display during submit:
    sys.stderr = None
    sys.stdout = None

    form_container = evt.target.closest("form")
    question_nbr = int(form_container.select(".editor-container")[0].attrs["data-editornbr"])
    editor = get_editor(question_nbr)

    question_data = globals()["data"][question_nbr]
    python_code = editor.getValue()
    request_obj = create_request(question_data, python_code)
    input_HTML = form_container.select('input[name="answer"]')
    input_HTML[0].attrs['value'] = javascript.JSON.stringify(request_obj)
    return True


def start_instance(editor_number):
    # document.select(".wait")[0].remove()
    document.select(".editor-container")[editor_number].style.visibility = "visible"
    display_only_mode = set_data(editor_number)

    # We need to know how many editors are in display only and already rendered
    # It will be used for index of events and console.
    numbers_editors_has_display_only = 0
    are_display_only = window.areDisplayOnly[:editor_number]
    for is_display_only in are_display_only:
        if is_display_only:
            numbers_editors_has_display_only += 1

    # no Python console if we are in display_only because not needed
    global consoles
    if not display_only_mode:
        # Create the interactive Python console
        console = Console(document.select(".console")[editor_number - numbers_editors_has_display_only])
        console.prompt_no_focus()
        consoles.append(console)
    else:
        consoles.append(None)
    create_editor(editor_number, display_only_mode, numbers_editors_has_display_only)
    window.savesFound = saves_found


def del_saves():
    global saves_found
    if not any(saves_found) and "saves" in storage:
        del storage["saves"]


def reset_instances():
    global consoles
    global editors
    global data
    consoles = []
    editors = []
    data = []


# We put the submit_code function outside the brython script
# To get the possibility of "preBind" it (see python_edit.js)
window.create_request_data = submit_code
# start_instance()

# For trigger after loading
window.reset_instances = reset_instances
window.start_instance = start_instance
window.toggle_lock = toggle_lock
window.hasDoneInit = True
window.delSaves = del_saves

# Start instances
containers = document.select("div[data-kwyk=python_editor]")
javascript.this().INIT_MULTIPLE_EDITORS(containers)